from functools import reduce
from operator import add, or_
from typing import Set

from sqllineage.core import LineageResult
from sqllineage.models import Table


class LineageCombiner:
    @staticmethod
    def combine(*args: LineageResult) -> LineageResult:
        raise NotImplementedError


class NaiveLineageCombiner(LineageCombiner):
    @staticmethod
    def combine(*args: LineageResult) -> LineageResult:
        return reduce(add, args, LineageResult())


class DefaultLineageCombiner(LineageCombiner):
    @staticmethod
    def combine(*args: LineageResult) -> LineageResult:
        combined_result = LineageResult()
        for lineage_result in args:
            if lineage_result.drop:
                for st_tables in (combined_result.read, combined_result.write):
                    st_tables -= lineage_result.drop
            elif lineage_result.rename:
                for (table_old, table_new) in lineage_result.rename:
                    for st_tables in (combined_result.read, combined_result.write):
                        if table_old in st_tables:
                            st_tables.remove(table_old)
                            st_tables.add(table_new)
            elif lineage_result.intermediate:
                combined_result.read |= (
                    lineage_result.read - lineage_result.intermediate
                )
                combined_result.write |= lineage_result.write
            else:
                combined_result.read |= lineage_result.read
                combined_result.write |= lineage_result.write
        combined_result.intermediate = combined_result.read.intersection(
            combined_result.write
        )
        self_depend_tables = reduce(
            or_,
            (
                lineage_result.read.intersection(lineage_result.write)
                for lineage_result in args
            ),
            set(),
        )  # type: Set[Table]
        combined_result.intermediate -= self_depend_tables
        combined_result.read -= combined_result.intermediate
        combined_result.write -= combined_result.intermediate
        return combined_result



import argparse
import importlib
import sys
from typing import List, Set, Type

import sqlparse
from sqlparse.sql import Statement

from sqllineage.combiners import DefaultLineageCombiner, LineageCombiner
from sqllineage.core import LineageAnalyzer
from sqllineage.models import Table


class LineageRunner(object):
    def __init__(
        self,
        sql: str,
        encoding: str = None,
        combiner: Type[LineageCombiner] = DefaultLineageCombiner,
        verbose: bool = False,
    ):
        self._encoding = encoding
        self._stmt = [
            s
            for s in sqlparse.parse(sql.strip(), self._encoding)
            if s.token_first(skip_cm=True)
        ]
        self._lineage_results = [LineageAnalyzer().analyze(stmt) for stmt in self._stmt]
        self._lineage_result = combiner.combine(*self._lineage_results)
        self._verbose = verbose

    def __str__(self):
        statements = self.statements(strip_comments=True)
        combined = """Statements(#): {stmt_cnt}
Source Tables:
    {source_tables}
Target Tables:
    {target_tables}
""".format(
            stmt_cnt=len(statements),
            source_tables="\n    ".join(str(t) for t in self.source_tables),
            target_tables="\n    ".join(str(t) for t in self.target_tables),
        )
        if self.intermediate_tables:
            combined += """Intermediate Tables:
    {intermediate_tables}""".format(
                intermediate_tables="\n    ".join(
                    str(t) for t in self.intermediate_tables
                )
            )
        if self._verbose:
            result = ""
            for i, lineage_result in enumerate(self._lineage_results):
                stmt_short = statements[i].replace("\n", "")
                if len(stmt_short) > 50:
                    stmt_short = stmt_short[:50] + "..."
                result += """Statement #{ord}: {stmt}
    {content}
""".format(
                    ord=i + 1,
                    content=str(lineage_result).replace("\n", "\n    "),
                    stmt=stmt_short,
                )
            combined = result + "==========\nSummary:\n" + combined
        return combined

    @property
    def statements_parsed(self) -> List[Statement]:
        return self._stmt

    def statements(self, **kwargs) -> List[str]:
        return [sqlparse.format(s.value, **kwargs) for s in self.statements_parsed]

    @property
    def source_tables(self) -> Set[Table]:
        return self._lineage_result.read

    @property
    def target_tables(self) -> Set[Table]:
        return self._lineage_result.write

    @property
    def intermediate_tables(self) -> Set[Table]:
        return self._lineage_result.intermediate


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="sqllineage", description="SQL Lineage Parser."
    )
    parser.add_argument(
        "-e", metavar="<quoted-query-string>", help="SQL from command line"
    )
    parser.add_argument("-f", metavar="<filename>", help="SQL from files")
    parser.add_argument(
        "-c",
        metavar="<combiner>",
        help="specify Combiner to combine lineage result from multiple statements, "
        "Default Value: sqllineage.combiners.DefaultLineageCombiner",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="increase output verbosity, show statement level lineage result",
        action="store_true",
    )
    args = parser.parse_args()
    combiner = DefaultLineageCombiner
    if args.c:
        try:
            combiner_module, combiner_name = args.c.rsplit(".", maxsplit=1)
            combiner = getattr(importlib.import_module(combiner_module), combiner_name)
        except ValueError:
            print(
                "ERROR: Combiner should be named as module.class, got {} instead".format(
                    args.c
                ),
                file=sys.stderr,
            )
            raise
        except (ImportError, AttributeError):
            print("ERROR: No such combiner: {}".format(args.c), file=sys.stderr)
            raise
    if args.e and args.f:
        print(
            "WARNING: Both -e and -f options are specified. -e option will be ignored",
            file=sys.stderr,
        )
    if args.f:
        try:
            with open(args.f) as f:
                sql = f.read()
            print(LineageRunner(sql, combiner=combiner, verbose=args.verbose))
        except FileNotFoundError:
            print("ERROR: No such file: {}".format(args.f), file=sys.stderr)
        except PermissionError:
            print(
                "ERROR: Permission denied when reading file '{}'".format(args.f),
                file=sys.stderr,
            )
    elif args.e:
        print(LineageRunner(args.e, combiner=combiner, verbose=args.verbose))
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



from setuptools import find_packages, setup

from sqllineage import name, version

with open("README.md", "r") as f:
    long_description = f.read()

ci_requires = [
    "bandit",
    "flake8",
    "flake8-blind-except",
    "flake8-builtins",
    "flake8-import-order",
    "flake8-logging-format",
    "mypy",
    "pytest>=4.5.0,<5.0",
    "pytest-cov",
    "tox>=3.11.0,<4.0",
    "twine",
    "wheel",
]

setup(
    name=name,
    version=version,
    author="Reata",
    author_email="reddevil.hjw@gmail.com",
    description="SQL Lineage Analysis Tool powered by Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/reata/sqllineage",
    packages=find_packages(exclude=("tests",)),
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    python_requires=">=3.5",
    install_requires=["sqlparse>=0.3.0,<0.4"],
    entry_points={"console_scripts": ["sqllineage = sqllineage.runner:main"]},
    extras_require={"ci": ci_requires},
)



import re
from typing import Set, TYPE_CHECKING, Tuple

from sqlparse.sql import (
    Comment,
    Comparison,
    Function,
    Identifier,
    Parenthesis,
    Statement,
    TokenList,
)
from sqlparse.tokens import Keyword, Token

from sqllineage.exceptions import SQLLineageException
from sqllineage.models import Table

SOURCE_TABLE_TOKENS = (
    r"FROM",
    # inspired by https://github.com/andialbrecht/sqlparse/blob/master/sqlparse/keywords.py
    r"((LEFT\s+|RIGHT\s+|FULL\s+)?(INNER\s+|OUTER\s+|STRAIGHT\s+)?|(CROSS\s+|NATURAL\s+)?)?JOIN",
)
TARGET_TABLE_TOKENS = ("INTO", "OVERWRITE", "TABLE", "VIEW")
TEMP_TABLE_TOKENS = ("WITH",)


class LineageResult:
    """Statement(s) Level Lineage Result."""

    __slots__ = ["read", "write", "rename", "drop", "intermediate"]
    if TYPE_CHECKING:
        read = write = drop = intermediate = set()  # type: Set[Table]
        rename = set()  # type: Set[Tuple[Table, Table]]

    def __init__(self) -> None:
        for attr in self.__slots__:
            setattr(self, attr, set())

    def __add__(self, other):
        lineage_result = LineageResult()
        for attr in self.__slots__:
            setattr(
                lineage_result, attr, getattr(self, attr).union(getattr(other, attr))
            )
        return lineage_result

    def __str__(self):
        return "\n".join(
            "table {attr}: {attr_val}".format(
                attr=attr, attr_val=getattr(self, attr) if getattr(self, attr) else "{}"
            )
            for attr in self.__slots__
        )

    def __repr__(self):
        return str(self)


class LineageAnalyzer:
    """SQL Statement Level Lineage Analyzer."""

    def __init__(self) -> None:
        self._lineage_result = LineageResult()

    def analyze(self, stmt: Statement) -> LineageResult:
        if stmt.get_type() == "DROP":
            self._extract_from_DDL_DROP(stmt)
        elif stmt.get_type() == "ALTER":
            self._extract_from_DDL_ALTER(stmt)
        elif (
            stmt.get_type() == "DELETE"
            or stmt.token_first(skip_cm=True).normalized == "TRUNCATE"
            or stmt.token_first(skip_cm=True).normalized.upper() == "REFRESH"
            or stmt.token_first(skip_cm=True).normalized == "CACHE"
        ):
            pass
        else:
            # DML parsing logic also applies to CREATE DDL
            self._extract_from_DML(stmt)
        return self._lineage_result

    def _extract_from_DDL_DROP(self, stmt: Statement) -> None:
        for table in {
            Table.create(t) for t in stmt.tokens if isinstance(t, Identifier)
        }:
            self._lineage_result.drop.add(table)

    def _extract_from_DDL_ALTER(self, stmt: Statement) -> None:
        tables = [Table.create(t) for t in stmt.tokens if isinstance(t, Identifier)]
        keywords = [t for t in stmt.tokens if t.ttype is Keyword]
        if any(k.normalized == "RENAME" for k in keywords) and len(tables) == 2:
            self._lineage_result.rename.add((tables[0], tables[1]))

    def _extract_from_DML(self, token: Token) -> None:
        source_table_token_flag = (
            target_table_token_flag
        ) = temp_table_token_flag = False
        for sub_token in token.tokens:
            if isinstance(sub_token, TokenList):
                self._extract_from_DML(sub_token)
            if sub_token.ttype in Keyword:
                if any(
                    re.match(regex, sub_token.normalized)
                    for regex in SOURCE_TABLE_TOKENS
                ):
                    source_table_token_flag = True
                elif sub_token.normalized in TARGET_TABLE_TOKENS:
                    target_table_token_flag = True
                elif sub_token.normalized in TEMP_TABLE_TOKENS:
                    temp_table_token_flag = True
                continue
            if source_table_token_flag:
                if self.__token_negligible_before_tablename(sub_token):
                    continue
                else:
                    if isinstance(sub_token, Identifier):
                        if isinstance(sub_token.token_first(skip_cm=True), Parenthesis):
                            # SELECT col1 FROM (SELECT col2 FROM tab1) dt, the subquery will be parsed as Identifier
                            # and this Identifier's get_real_name method would return alias name dt
                            # referring https://github.com/andialbrecht/sqlparse/issues/218 for further information
                            pass
                        else:
                            self._lineage_result.read.add(Table.create(sub_token))
                    elif isinstance(sub_token, Parenthesis):
                        # SELECT col1 FROM (SELECT col2 FROM tab1), the subquery will be parsed as Parenthesis
                        # This syntax without alias for subquery is invalid in MySQL, while valid for SparkSQL
                        pass
                    else:
                        raise SQLLineageException(
                            "An Identifier is expected, got %s[value: %s] instead"
                            % (type(sub_token).__name__, sub_token)
                        )
                    source_table_token_flag = False
            elif target_table_token_flag:
                if self.__token_negligible_before_tablename(sub_token):
                    continue
                elif isinstance(sub_token, Function):
                    # insert into tab (col1, col2) values (val1, val2); Here tab (col1, col2) will be parsed as Function
                    # referring https://github.com/andialbrecht/sqlparse/issues/483 for further information
                    if not isinstance(sub_token.token_first(skip_cm=True), Identifier):
                        raise SQLLineageException(
                            "An Identifier is expected, got %s[value: %s] instead"
                            % (type(sub_token).__name__, sub_token)
                        )
                    self._lineage_result.write.add(
                        Table.create(sub_token.token_first(skip_cm=True))
                    )
                elif isinstance(sub_token, Comparison):
                    # create table tab1 like tab2, tab1 like tab2 will be parsed as Comparison
                    # referring https://github.com/andialbrecht/sqlparse/issues/543 for further information
                    if not (
                        isinstance(sub_token.left, Identifier)
                        and isinstance(sub_token.right, Identifier)
                    ):
                        raise SQLLineageException(
                            "An Identifier is expected, got %s[value: %s] instead"
                            % (type(sub_token).__name__, sub_token)
                        )
                    self._lineage_result.write.add(Table.create(sub_token.left))
                    self._lineage_result.read.add(Table.create(sub_token.right))
                else:
                    if not isinstance(sub_token, Identifier):
                        raise SQLLineageException(
                            "An Identifier is expected, got %s[value: %s] instead"
                            % (type(sub_token).__name__, sub_token)
                        )
                    self._lineage_result.write.add(Table.create(sub_token))
                target_table_token_flag = False
            elif temp_table_token_flag:
                if self.__token_negligible_before_tablename(sub_token):
                    continue
                else:
                    if not isinstance(sub_token, Identifier):
                        raise SQLLineageException(
                            "An Identifier is expected, got %s[value: %s] instead"
                            % (type(sub_token).__name__, sub_token)
                        )
                    self._lineage_result.intermediate.add(Table.create(sub_token))
                    self._extract_from_DML(sub_token)
                    temp_table_token_flag = False

    @classmethod
    def __token_negligible_before_tablename(cls, token: Token) -> bool:
        return token.is_whitespace or isinstance(token, Comment)
